<html xmlns="http://www.w3.org/1999/xhtml">

  <head>
    <meta content="text/html; charset=utf-8" http-equiv="Content-Type" />

    <link href="http://fonts.googleapis.com/css?family=Roboto+Condensed:400italic,700italic,700,400" rel="stylesheet" type="text/css" />
    <link href="http://fonts.googleapis.com/css?family=Roboto:400italic,700italic,700,400" rel="stylesheet" type="text/css" />
    <link href='http://fonts.googleapis.com/css?family=Droid+Sans+Mono' rel='stylesheet' type='text/css' />

    <link href="/resources/web.css" rel="stylesheet" type="text/css" />
    <title>{{title}}: {{pagetitle}}</title>

    <!-- begin scripts -->
    {{!scripts}}
    <!-- end scripts -->
  </head>

  <body>

    <!-- The title of the input file, from the 'title' option. -->
    <div id="title">{{title}}{{view_title}}</div>

    {{!overlay}}

    <div class="navigation">
      {{!navigation}}
    </div>

    <!-- The title of this page, what information we're displaying. -->
    <h1 id="pagetitle">{{pagetitle}}</h1>

    <!-- The contents of this page. -->
    <div id="contents">
      {{!contents}}
    </div>

  </body>
</html>
